@@ -1,5 +1,6 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
+from collections import defaultdict
 import time
 import logging
 import os
@@ -81,7 +82,7 @@ class TryAdvisoryLock(Function):
 class Database(DatabaseInterface):
 
     _lock = RLock()
-    _databases = {}
+    _databases = defaultdict(dict)
     _connpool = None
     _list_cache = {}
     _list_cache_timestamp = {}
@@ -103,13 +104,14 @@ class Database(DatabaseInterface):
     def __new__(cls, name='template1'):
         with cls._lock:
             now = datetime.now()
-            for database in list(cls._databases.values()):
+            databases = cls._databases[os.getpid()]
+            for database in list(databases.values()):
                 if ((now - database._last_use).total_seconds() > _timeout
                         and database.name != name
                         and not database._connpool._used):
                     database.close()
-            if name in cls._databases:
-                inst = cls._databases[name]
+            if name in databases:
+                inst = databases[name]
             else:
                 if name == 'template1':
                     minconn = 0
@@ -118,9 +120,10 @@ class Database(DatabaseInterface):
                 inst = DatabaseInterface.__new__(cls, name=name)
                 logger.info('connect to "%s"', name)
                 inst._connpool = ThreadedConnectionPool(
-                    minconn, _maxconn, cls.dsn(name),
-                    cursor_factory=LoggingCursor)
-                cls._databases[name] = inst
+                    minconn, _maxconn,
+                    cursor_factory=LoggingCursor,
+                    **cls._connection_params(name))
+                databases[name] = inst
             inst._last_use = datetime.now()
             return inst
 
@@ -128,15 +131,20 @@ class Database(DatabaseInterface):
         super(Database, self).__init__(name)
 
     @classmethod
-    def dsn(cls, name):
+    def _connection_params(cls, name):
         uri = parse_uri(config.get('database', 'uri'))
-        host = uri.hostname and "host=%s" % uri.hostname or ''
-        port = uri.port and "port=%s" % uri.port or ''
-        name = "dbname=%s" % name
-        user = uri.username and "user=%s" % uri.username or ''
-        password = ("password=%s" % urllib.parse.unquote_plus(uri.password)
-            if uri.password else '')
-        return '%s %s %s %s %s' % (host, port, name, user, password)
+        params = {
+            'dbname': name,
+            }
+        if uri.username:
+            params['user'] = uri.username
+        if uri.password:
+            params['password'] = urllib.parse.unquote_plus(uri.password)
+        if uri.hostname:
+            params['host'] = uri.hostname
+        if uri.port:
+            params['port'] = uri.port
+        return params
 
     def connect(self):
         return self
@@ -152,11 +160,17 @@ class Database(DatabaseInterface):
                     time.sleep(1)
                     continue
                 raise
+        # We do not use set_session because psycopg2 < 2.7 and psycopg2cffi
+        # change the default_transaction_* attributes which breaks external
+        # pooling at the transaction level.
         if autocommit:
             conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
         else:
             conn.set_isolation_level(ISOLATION_LEVEL_REPEATABLE_READ)
-        if readonly:
+        # psycopg2cffi does not have the readonly property
+        if hasattr(conn, 'readonly'):
+            conn.readonly = readonly
+        elif not autocommit and readonly:
             cursor = conn.cursor()
             cursor.execute('SET TRANSACTION READ ONLY')
         return conn
@@ -168,7 +182,7 @@ class Database(DatabaseInterface):
         with self._lock:
             logger.info('disconnect from "%s"', self.name)
             self._connpool.closeall()
-            self._databases.pop(self.name)
+            self._databases[os.getpid()].pop(self.name)
 
     @classmethod
     def create(cls, connection, database_name, template='template0'):
@@ -205,7 +219,8 @@ class Database(DatabaseInterface):
             res = []
             for db_name, in cursor:
                 try:
-                    with connect(self.dsn(db_name)) as conn:
+                    with connect(**self._connection_params(db_name)
+                            ) as conn:
                         if self._test(conn, hostname=hostname):
                             res.append(db_name)
                 except Exception:
